# SPDX-FileCopyrightText: ☭ Emery Hemingway
# SPDX-License-Identifier: Unlicense

## ESC/P printer control actor.

import
  std/[cmdline, oserrors, posix, sets],
  preserves, preserves/sugar,
  syndicate, syndicate/relays,
  ./private/esc_p

proc echo(args: varargs[string, `$`]) {.used.} =
  stderr.writeLine(args)

type
  HandleSet = HashSet[Handle]

  Printer = ref object of Entity
    device: cint
    boldHandles, italicHandles, superscriptHandles, subscriptHandles: HandleSet

proc write(printer: Printer; s: string) {.inline.} =
  if posix.write(printer.device, s[0].addr, s.len) < 0:
    osLastError().osErrorMsg().quit()

proc writeLine(printer: Printer; s: string) {.inline.} =
  printer.write(s)
  printer.write("\r\n")

method message(printer: Printer; t: Turn; a: AssertionRef) =
  if a.value.isString:
    printer.write(a.value.string)
      # TODO: unicode?
      # TODO: line breaks?

proc assert(printer: Printer; handles: var HandleSet; ctrl: string; h: Handle) =
  if handles.len == 0: printer.write(ctrl)
  handles.incl h

proc retract(printer: Printer; handles: var HandleSet; ctrl: string; h: Handle) =
  handles.excl h
  if handles.len == 0: printer.write(ctrl)

method publish(printer: Printer; t: Turn; a: AssertionRef; h: Handle) =
  if a.value.isRecord("bold"):
    printer.assert(printer.boldHandles, SelectBoldFont, h)

  elif a.value.isRecord("italic"):
    printer.assert(printer.italicHandles, SelectItalicFont, h)

  elif a.value.isRecord("superscript"):
    printer.assert(printer.superscriptHandles, SelectSuperScript, h)

  elif a.value.isRecord("subscript"):
    printer.assert(printer.subscriptHandles, SelectSubScript, h)

method retract(printer: Printer; t: Turn; h: Handle) =
  if printer.boldHandles.contains h:
    printer.retract(printer.boldHandles, CancelBoldFont, h)

  elif printer.italicHandles.contains h:
    printer.retract(printer.italicHandles, CanceItalicFont, h)

  elif printer.superscriptHandles.contains h:
    printer.retract(printer.superscriptHandles, CancelAltScript, h)

  elif printer.subscriptHandles.contains h:
    printer.retract(printer.subscriptHandles, CancelAltScript, h)

proc openPrinter(turn: Turn; devicePath: string): Printer =
  new result
  result.facet = turn.facet
  result.device = posix.open(devicePath, O_WRONLY, 0)
  result.write(InitializePrinter)

proc main =
  let devicePath = paramStr(1)
  runActor(devicePath) do (turn: Turn):
    let printer = turn.newCap openPrinter(turn, devicePath)
    resolveEnvironment(turn) do (turn: Turn; ds: Cap):
      publish(turn, ds, initRecord(
        toSymbol"printer", printer.embed, %devicePath))

main()
